/* $Id: vircam_science_postprocess.c,v 1.2 2013-10-21 14:43:04 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-21 14:43:04 $
 * $Revision: 1.2 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <cpl.h>
#include <math.h>

#include <casu_utils.h>
#include <casu_mods.h>
#include <casu_mask.h>
#include <casu_sky.h>
#include <casu_filt.h>
#include <casu_wcsutils.h>
#include <casu_stats.h>

#include "vircam_utils.h"
#include "vircam_pfits.h"
#include "vircam_dfs.h"
#include "vircam_mods.h"
#include "vircam_paf.h"
#include "vircam_sky.h"


struct {

    /* Input command line parameters */

    int nebulise;
    int grout;
    int minphotom;
    int prettynames;
    int cdssearch;
    int neb_medfilt;
    int neb_linfilt;
    int cat_ipix;
    float cat_thresh;
    int cat_icrowd;
    float cat_rcore;
    int cat_nbsize;
    char *cacheloc;
    float magerrcut; 
} vircam_sci_postproc_config;

static char vircam_recipename[VIRCAM_PATHSZ];
static char vircam_recipepaf[VIRCAM_PATHSZ];

struct {
    cpl_size      *labels;
    cpl_frameset  *paws;
    cpl_frameset  *confs;
    cpl_frameset  *cats;

    cpl_frame     *phottab;
    cpl_table     *tphottab;
    cpl_frame     *catindex;
    char          *catpath;
    char          *catname;
    cpl_frame     *schlf_n;
    cpl_frame     *schlf_s;

    casu_fits     *outmos;
    casu_fits     *outmosconf;
    casu_tfits    *outmoscat;
} ps;

/* Main stuff for CPL/esorex */

static int vircam_science_postprocess_create(cpl_plugin *plugin);
static int vircam_science_postprocess_exec(cpl_plugin *plugin);
static int vircam_science_postprocess_destroy(cpl_plugin *plugin);
static int vircam_science_postprocess(cpl_parameterlist *parlist,
                                      cpl_frameset *framelist);


static int vircam_sci_postproc_save_image(casu_fits *outim, 
                                          cpl_frameset *framelist,
                                          cpl_parameterlist *parlist,
                                          cpl_frame *template, int isim);
static int vircam_sci_postproc_save_cat(casu_tfits *outcat, 
                                        cpl_frameset *framelist,
                                        cpl_parameterlist *parlist,
                                        cpl_frame *template);
static void vircam_sci_postproc_init(void);
static int vircam_sci_testfrm_1(cpl_frame *fr, int nextn_expected, int isimg);
static void vircam_sci_product_name(const char *template, int producttype, 
                                    int nametype, int fnumber, char *outfname);
static void vircam_sci_postproc_tidy(void);

static char vircam_science_postprocess_description[] =
"vircam_science_postprocess -- VIRCAM science postprocess recipe.\n\n"
"Mosaic a list of vircam stacks and generate a catalogue. Optionally nebulise\n"
"before generating the catalogue.\n"
"The program accepts the following files in the SOF:\n\n"
"    Tag                    Description\n"
"    -----------------------------------------------------------------------\n"
"    %-22s A list of science jittered stacks images\n"
"    %-22s A list of science jittered stack confidence maps\n"
"    %-22s A list of science jittered stack catalogues\n"
"    %-22s Northern Schlegel Map\n"
"    %-22s Southern Schlegel Map\n"
"    %-22s A master 2MASS index\n"
"    %-22s A photometric calibration table\n"
"All of the above are required\n"
"\n";

/**@{*/

/**
    \ingroup recipelist
    \defgroup vircam_science_postprocess vircam_science_postprocess
    \brief Tile a set of vircam stacks

    \par Name: 
        vircam_science_postprocess
    \par Purpose: 
        Tile a complete set of vircam stacks
    \par Description: 
        The stacks from a vircam tile are combined into a single stack.
        A source catalogue is generated and photometric and astrometric
        calibration is done relative to 2MASS.
    \par Language:
        C
    \par Parameters:
        General reduction parameters:
        - \b nebulise (bool): If set, then the catalogues are derived from a 
             tile that has been corrected with an unsharp mask
        - \b grout (bool): If set then the source catalogue will be corrected
             using the grout algorithm
        - \b minphotom (int): The minimum number of standards for photometry
        - \b prettynames (bool): True if we're using nice file names
        - \b cdssearch (bool): True if we get 2MASS standards from CDS.
        - \b cacheloc (string): The location of the standard star cache
        - \b magerrcut (float): A cut in the magnitude error for photometric
             standard stars
        
        Parameters for source detection on the tile:
        - \b cat_ipix (int): The minimum allowable size of an object
        - \b cat_thresh (float): The detection threshold in sigma above sky
        - \b cat_icrowd (int): If set then the deblending software will be used
        - \b cat_rcore (float): The core radius in pixels
        - \b cat_nbsize (int): The smoothing box size for background map 
             estimation
        
        Parameters for nebuliser algorithm:
        - \b neb_medfilt (int): Cell size for the median filter
        - \b neb_linfilt (int): Cell size for the linear filter
        
    \par Input File Types:
        The following list gives the file types that can appear in the SOF
        file. The word in bold is the DO category value.
        - \b JITTERED_IMAGE_SCI (required): The list of science jittered images.
        - \b CONFIDENCE_MAP_SCI (required): The list of science jittered confidence maps
        - \b OBJECT_CATALOGUE_SCI (required): The list of science source catalogues
        - \b PHOTCAL_TAB (required): A photometric calibration table
        - \b MASTER_2MASS_CATLAOGUE (required): A master 2MASS catalogue
             location (unless using cdssearch). 
        - \b SCHLEGEL_MAP_NORTH (required): The Northern Schlegel dust map
        - \b SCHLEGEL_MAP_SOUTH (required): The Southern Schlegel dust map
    \par Output Products:
        The following list gives the output data products that are generated
        by this recipe. The word in bold gives the DPR CATG keyword value for
        each product:
        - The output mosaicked tile (\b TILED_IMAGE)
        - The output mosaicked tile confidence map (\b TILED_CONFIDENCE_MAP)
        - The extracted source catalogue for the tile (\b OBJECT_CATALOGUE_SCI)
    \par Output QC Parameters:
        - \b SATURATION
             The saturation level in ADUs.
        - \b MEAN_SKY
             The mean level of the background sky over the image in ADUs.
        - \b SKY_NOISE
             The RMS of the background sky over the image in ADUs
        - \b IMAGE_SIZE
             The average size of stellar objects on the image in pixels
        - \b ELLIPTICITY
             The average ellipticity of stellar objects on the image
        - \b POSANG
             The average position angle in degrees from North towards East.
             NB: this value only makes sense if the ellipticity is significant
        - \b APERTURE_CORR
             The aperture correction for an aperture of radius rcore.
        - \b NOISE_OBJ
             The number of noise objects found in the image
        - \b MAGZPT
             The photometric zero point
        - \b MAGZERR
             The internal error in the photometric zero point
        - \b MAGNZPT
             The number of stars used to determine the magnitude zero point
        - \b MAGNCUT
             The number of stars cut from magnitude zero point calculation
        - \b SKYBRIGHT
             The sky brightness in mag/arcsec**2
        - \b LIMITING_MAG
             The limiting magnitude for this image for a 5 sigma detection.
        - \b WCS_DCRVAL1
             The offset of the equatorial coordinate represented by CRVAL1 
             from the raw frame to the reduced one (degrees).
        - \b WCS_DCRVAL2
             The offset of the equatorial coordinate represented by CRVAL2
             from the raw frame to the reduced one (degrees).
        - \b WCS_DTHETA
             The change in the WCS coordinate system rotation from the raw
             to the reduced frame (degrees)
        - \b WCS_SCALE
             The scale of the pixels in the reduced frames in arcseconds per
             pixel
        - \b WCS_SHEAR
             The shear of the astrometric solution in the form of the 
             difference between the rotation of the x solution and the rotation
             of the y solution (abs(xrot) - abs(yrot) in degrees)
        - \b WCS_RMS
             The average error in the WCS fit (arcsec)
    \par Notes
        None.
    \par Fatal Error Conditions:
        - NULL input frameset
        - Input frameset headers incorrect meaning that RAW and CALIB frame
          can't be distinguished
        - No science frames in the input frameset
        - Required master calibration images and tables are missing
        - Unable to save data products
    \par Non-Fatal Error Conditions:
        - None
    \par Conditions Leading To Dummy Products:
        - Master calibration images either won't load or are flagged as dummy
        - The detector for the current image extension is flagged dead in
          all science frames
        - Various processing routines fail.
    \par Author:
        Jim Lewis, CASU
    \par Code Reference: 
        vircam_science_postprocess.c
*/

/* Function code */

/*---------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*---------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
    cpl_recipe  *recipe = cpl_calloc(1,sizeof(*recipe));
    cpl_plugin  *plugin = &recipe->interface;
    char alldesc[SZ_ALLDESC];
    (void)snprintf(alldesc,SZ_ALLDESC,
                   vircam_science_postprocess_description,
                   VIRCAM_PRO_JITTERED_SCI,VIRCAM_PRO_CONF_SCI,
                   VIRCAM_PRO_OBJCAT_SCI,VIRCAM_CAL_SCHL_N,
                   VIRCAM_CAL_SCHL_S,VIRCAM_CAL_2MASS,VIRCAM_CAL_PHOTTAB);

    cpl_plugin_init(plugin, 
                    CPL_PLUGIN_API,
                    VIRCAM_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "vircam_science_postprocess",
                    "VIRCAM science postprocessing recipe",
                    alldesc,
                    "Jim Lewis",
                    "jrl@ast.cam.ac.uk",
                    vircam_get_license(),
                    vircam_science_postprocess_create,
                    vircam_science_postprocess_exec,
                    vircam_science_postprocess_destroy);

    cpl_pluginlist_append(list,plugin);

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_postprocess_create(cpl_plugin *plugin) {
    cpl_recipe      *recipe;
    cpl_parameter   *p;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    /* Create the parameters list in the cpl_recipe object */

    recipe->parameters = cpl_parameterlist_new();

    /* Fill in flag to use the nebuliser */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.nebulise",
                                CPL_TYPE_BOOL,
                                "Nebulise the stacks before object detection?",
                                "vircam.vircam_science_postprocess",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"nebulise");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to grout the output catalogue */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.grout",
                                CPL_TYPE_BOOL,
                                "Grout the output tile catalogue?",
                                "vircam.vircam_science_postprocess",TRUE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"grout");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the minimum number of stars for photometry */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.minphotom",
                                CPL_TYPE_INT,
                                "Minimum number of stars for photometry solution",
                                "vircam.vircam_science_postprocess",20,1,
                                100000);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"minphotom");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to use pretty names */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.prettynames",
                                CPL_TYPE_BOOL,
                                "Use pretty names?",
                                "vircam.vircam_science_postprocess",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"prettynames");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to decide whether to use CDS to get 2MASS standards */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.cdssearch",
                                CPL_TYPE_BOOL,"Use CDS for 2MASS standards?",
                                "vircam.vircam_science_postprocess",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cdssearch");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the median filter size for nebuliser */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.neb_medfilt",
                                CPL_TYPE_INT,
                                "Median filter size for nebuliser",
                                "vircam.vircam_science_postprocess",101,11,
                                2047);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"neb_medfilt");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the linear filter size for nebuliser */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.neb_linfilt",
                                CPL_TYPE_INT,
                                "Linear filter size for nebuliser",
                                "vircam.vircam_science_postprocess",33,3,2047);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"neb_linfilt");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the minimum object size */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.cat_ipix",
                                CPL_TYPE_INT,
                                "Minimum pixel area for each detected object",
                                "vircam.vircam_science_postprocess",4,1,100000);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cat_ipix");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the detection threshold parameter */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.cat_thresh",
                                CPL_TYPE_DOUBLE,
                                "Detection threshold in sigma above sky",
                                "vircam.vircam_science_postprocess",1.25,
                                1.0e-6,1.0e10);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cat_thresh");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to use deblending software or not */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.cat_icrowd",
                                CPL_TYPE_BOOL,"Use deblending?",
                                "vircam.vircam_science_postprocess",TRUE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cat_icrowd");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in core radius */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.cat_rcore",
                                CPL_TYPE_DOUBLE,"Value of Rcore in pixels",
                                "vircam.vircam_science_postprocess",3.0,1.0e-6,
                                1024.0);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cat_rcore");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in background smoothing box size */

    p = cpl_parameter_new_range("vircam.vircam_science_postprocess.cat_nbsize",
                                CPL_TYPE_INT,"Background smoothing box size",
                                "vircam.vircam_science_postprocess",64,1,2048);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cat_nbsize");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Location for standard star cache */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.cacheloc",
                                CPL_TYPE_STRING,
                                "Location for standard star cache",
                                "vircam.vircam_science_postprocess",".");
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cacheloc");
    cpl_parameterlist_append(recipe->parameters,p);  

    /* Magnitude error cut */

    p = cpl_parameter_new_value("vircam.vircam_science_postprocess.magerrcut",
                                CPL_TYPE_DOUBLE,
                                "Magnitude error cut",
                                "vircam.vircam_science_postprocess",100.0);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"magerrcut");
    cpl_parameterlist_append(recipe->parameters,p);  

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_postprocess_exec(cpl_plugin *plugin) {
    cpl_recipe  *recipe;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    return(vircam_science_postprocess(recipe->parameters,recipe->frames));
}
                                
/*---------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_postprocess_destroy(cpl_plugin *plugin) {
    cpl_recipe *recipe ;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    cpl_parameterlist_delete(recipe->parameters);
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_postprocess(cpl_parameterlist *parlist,
                                      cpl_frameset *framelist) {
    const char *fctid="vircam_scienc_postprocess";
    cpl_parameter *p;
    int nfail,status,n,i,j;
    cpl_size nlab;
    cpl_frameset *tmpframeset;
    cpl_frame *frm,*frmc,*newfrm;
    char tmpfilename[VIRCAM_PATHSZ],filt[16];
    casu_fits *inf,*infc,*tmpoutmos,*tmpoutmosconf,*backmap;
    casu_tfits *outmoscatg,*matchstds;
    cpl_table *stdscat,*tab;
    float gain;

    /* Check validity of the input frameset */

    if (framelist == NULL || cpl_frameset_get_size(framelist) <= 0) {
        cpl_msg_error(fctid,"Input framelist NULL or has no input data");
        return(-1);
    }

    /* Initialise some things */

    vircam_sci_postproc_init();
    (void)strncpy(vircam_recipename,fctid,VIRCAM_PATHSZ);
    (void)snprintf(vircam_recipepaf,VIRCAM_PATHSZ,"VIRCAM/%s",fctid);

    /* Stacking special parameters */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.nebulise");
    vircam_sci_postproc_config.nebulise = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.grout");
    vircam_sci_postproc_config.grout = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.minphotom");
    vircam_sci_postproc_config.minphotom = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.prettynames");
    vircam_sci_postproc_config.prettynames = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cdssearch");
    vircam_sci_postproc_config.cdssearch = (cpl_parameter_get_bool(p) ? 1 : 0);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.neb_medfilt");
    vircam_sci_postproc_config.neb_medfilt = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.neb_linfilt");
    vircam_sci_postproc_config.neb_linfilt = cpl_parameter_get_int(p);

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cat_ipix");
    vircam_sci_postproc_config.cat_ipix = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cat_thresh");
    vircam_sci_postproc_config.cat_thresh = (float)cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cat_icrowd");
    vircam_sci_postproc_config.cat_icrowd = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cat_rcore");
    vircam_sci_postproc_config.cat_rcore = (float)cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cat_nbsize");
    vircam_sci_postproc_config.cat_nbsize = cpl_parameter_get_int(p);

    /* Cache location */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.cacheloc");
    vircam_sci_postproc_config.cacheloc = (char *)cpl_parameter_get_string(p);

    /* Magnitude error cut */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_postprocess.magerrcut");
    vircam_sci_postproc_config.magerrcut = (float)cpl_parameter_get_double(p);

    /* Sort out raw from calib frames */

    if (vircam_dfs_set_groups(framelist) != CASU_OK) {
        cpl_msg_error(fctid,"Cannot identify RAW and CALIB frames");
        vircam_sci_postproc_tidy();
        return(-1);
    }

    /* Label the input frames */

    if ((ps.labels = cpl_frameset_labelise(framelist,casu_compare_tags,
                                           &nlab)) == NULL) {
        cpl_msg_error(fctid,"Cannot labelise the input frames");
        vircam_sci_postproc_tidy();
        return(-1);
    }

    /* Get the input science frames and their confidence maps */

    if ((ps.paws = 
         casu_frameset_subgroup(framelist,ps.labels,(int)nlab,
                                  VIRCAM_PRO_JITTERED_SCI)) == NULL) {
        cpl_msg_error(fctid,"No science images to process!");
        vircam_sci_postproc_tidy();
        return(-1);
    }
    if ((ps.confs = 
         casu_frameset_subgroup(framelist,ps.labels,(int)nlab,
                                  VIRCAM_PRO_CONF_SCI)) == NULL) {
        cpl_msg_error(fctid,"No science confidence maps to process!");
        vircam_sci_postproc_tidy();
        return(-1);
    }
    if ((ps.cats = 
         casu_frameset_subgroup(framelist,ps.labels,(int)nlab,
                                  VIRCAM_PRO_OBJCAT_SCI)) == NULL) {
        cpl_msg_error(fctid,"No science object catalogues to process!");
        vircam_sci_postproc_tidy();
        return(-1);
    }

    /* Check to see if there is a photometric table */

    nfail = 0;
    if ((ps.phottab = casu_frameset_subgroup_1(framelist,ps.labels,(int)nlab,
                                                 VIRCAM_CAL_PHOTTAB)) == NULL) {
        cpl_msg_error(fctid,"No photometric table found");
        vircam_sci_postproc_tidy();
        return(-1);
    }
    n = cpl_frame_get_nextensions(ps.phottab);
    nfail += vircam_sci_testfrm_1(ps.phottab,n,0);


    /* Is the 2mass index file specified? */

    if (vircam_sci_postproc_config.cdssearch == 0) {
        if ((ps.catindex = casu_frameset_subgroup_1(framelist,ps.labels,(int)nlab,
                                                    VIRCAM_CAL_2MASS)) == NULL) {
            cpl_msg_error(fctid,"No 2MASS index found -- cannot continue");
            vircam_sci_postproc_tidy();
            return(-1);
        }
        nfail += vircam_sci_testfrm_1(ps.catindex,1,0);
    }

    /* Is the Northern Schlegel file specified? */

    if ((ps.schlf_n = casu_frameset_subgroup_1(framelist,ps.labels,(int)nlab,
                                                 VIRCAM_CAL_SCHL_N)) == NULL) {
        cpl_msg_error(fctid,"Schlegel North map not found -- cannot continue");
        vircam_sci_postproc_tidy();
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.schlf_n,0,1);

    /* Is the Southern Schlegel file specified? */

    if ((ps.schlf_s = casu_frameset_subgroup_1(framelist,ps.labels,(int)nlab,
                                                 VIRCAM_CAL_SCHL_S)) == NULL) {
        cpl_msg_error(fctid,"Schlegel South map not found -- cannot continue");
        vircam_sci_postproc_tidy();
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.schlf_s,0,1);

    /* Check the cache location is writable */

    if (access(vircam_sci_postproc_config.cacheloc,R_OK+W_OK+X_OK) != 0) {
        cpl_msg_error(fctid,"Cache location %s inacessible",
                      vircam_sci_postproc_config.cacheloc);
        nfail++;
    }

    /* Ok if any of this failed, then get out of here. This makes it a bit 
       simpler later on since we now know that all of the specified files 
       have the correct number of extensions and will load properly */

    if (nfail > 0) {
        cpl_msg_error(fctid,
                      "There are %" CPL_SIZE_FORMAT " input file errors -- cannot continue",
                      (cpl_size)nfail);
        vircam_sci_postproc_tidy();
        return(-1);
    }
    
    /* Get catalogue parameters */

    if (ps.catindex != NULL) {
        if (casu_catpars(ps.catindex,&(ps.catpath),&(ps.catname)) == CASU_FATAL) {
            vircam_sci_postproc_tidy();
            return(-1);
        }
        freeframe(ps.catindex);
    }

    /* Using the nebuliser before doing the catalogue, then create a frameset
       of nebulised pawprints */

    status = CASU_OK;
    if (vircam_sci_postproc_config.nebulise) {
        tmpframeset = cpl_frameset_new();
        n = (int)cpl_frameset_get_size(ps.paws);
        cpl_msg_info(fctid,"Nebulising pawprints");
        cpl_msg_indent_more();
        for (i = 0; i < n; i++) {
            cpl_msg_info(fctid,"Doing pawprint %" CPL_SIZE_FORMAT "",
                         (cpl_size)(i+1));
            frm = cpl_frameset_get_position(ps.paws,(cpl_size)i);
            frmc = cpl_frameset_get_position(ps.confs,(cpl_size)i);
            (void)sprintf(tmpfilename,"tmp_%s",basename((char *)cpl_frame_get_filename(frm)));
            newfrm = cpl_frame_new();
            cpl_frame_set_tag(newfrm,VIRCAM_SCI_OBJECT_RAW);
            cpl_frame_set_filename(newfrm,tmpfilename);
            cpl_frameset_insert(tmpframeset,newfrm);
            for (j = 1; j <= VIRCAM_NEXTN; j++) {
                inf = casu_fits_load(frm,CPL_TYPE_FLOAT,j);
                infc = casu_fits_load(frmc,CPL_TYPE_INT,j);

                (void)casu_nebuliser(inf,infc,101,33,3,1,0,0,0,0,10.0,3.0,
                                     &backmap,&status);
                if (status != CASU_OK) {
                    vircam_sci_postproc_tidy();
                    freefits(inf);
                    freefits(infc);
                    freeframeset(tmpframeset);
                    return(-1);
                }
                if (j == 1) {
                    if (access(tmpfilename,F_OK)) 
                        remove(tmpfilename);
                    cpl_image_save(NULL,tmpfilename,CPL_TYPE_UCHAR,
                                   casu_fits_get_phu(inf),CPL_IO_DEFAULT);
                }
                cpl_image_save(casu_fits_get_image(inf),tmpfilename,
                               CPL_TYPE_FLOAT,casu_fits_get_ehu(inf),
                               CPL_IO_EXTEND);
                freefits(inf);
                freefits(infc);
            }
        }
        cpl_msg_indent_less();

        /* Mosaic the nebulised files */

        cpl_msg_info(fctid,"Tiling nebulised pawprints");
        (void)vircam_mosaic(tmpframeset,ps.confs,1,2,0.0,"EXPTIME",
                            25,&tmpoutmos,&tmpoutmosconf,&status);
        if (status != CASU_OK) {
            vircam_sci_postproc_tidy();
            freeframeset(tmpframeset);
            freefits(tmpoutmos);
            freefits(tmpoutmosconf);
            return(-1);
        }

        /* Chuck nebulised pawprints */

        for (i = 0; i < n; i++) {
            frm = cpl_frameset_get_position(tmpframeset,i);
            remove(cpl_frame_get_filename(frm));
        }
        freeframeset(tmpframeset);

        /* Now do an imcore to create a catalogue from the nebulised mosaic. */

        cpl_msg_info(fctid,"Creating catalogue from nebulised mosaic");
        (void)vircam_pfits_get_gain(casu_fits_get_ehu(tmpoutmos),&gain);
        (void)casu_imcore(tmpoutmos,tmpoutmosconf,
                          vircam_sci_postproc_config.cat_ipix,
                          vircam_sci_postproc_config.cat_thresh,
                          vircam_sci_postproc_config.cat_icrowd,
                          vircam_sci_postproc_config.cat_rcore,
                          vircam_sci_postproc_config.cat_nbsize,6,
                          2.0,&(ps.outmoscat),gain,&status);
        if (status != CASU_OK) {
            vircam_sci_postproc_tidy();
            freeframeset(tmpframeset);
            freefits(tmpoutmos);
            freefits(tmpoutmosconf);
            return(-1);
        }

        /* Chuck the nebulised tile and confidence map */

        freefits(tmpoutmos);
        freefits(tmpoutmosconf);

        /* Do the real mosaic */
        
        cpl_msg_info(fctid,"Creating the real mosaic");
        (void)vircam_mosaic(ps.paws,ps.confs,1,2,0.0,"EXPTIME",25,&(ps.outmos),
                            &(ps.outmosconf),&status);
        if (status != CASU_OK) {
            vircam_sci_postproc_tidy();
            return(-1);
        }

    /* Not using the nebuliser, so just do an imcore to create a catalogue 
       from the real mosaic */

    } else {

        /* First do the real mosaic */
        
        cpl_msg_info(fctid,"Creating the real mosaic");
        (void)vircam_mosaic(ps.paws,ps.confs,1,2,0.0,"EXPTIME",25,&(ps.outmos),
                            &(ps.outmosconf),&status);
        if (status != CASU_OK) {
            vircam_sci_postproc_tidy();
            return(-1);
        }

        /* Now the imcore */

        cpl_msg_info(fctid,"Creating catalogue from mosaic");
        (void)vircam_pfits_get_gain(casu_fits_get_ehu(ps.outmos),&gain);
        (void)casu_imcore(ps.outmos,ps.outmosconf,
                          vircam_sci_postproc_config.cat_ipix,
                          vircam_sci_postproc_config.cat_thresh,
                          vircam_sci_postproc_config.cat_icrowd,
                          vircam_sci_postproc_config.cat_rcore,
                          vircam_sci_postproc_config.cat_nbsize,6,
                          2.0,&(ps.outmoscat),gain,&status);
    }

    /* Get some standard stars and match it against the catalogue. Do
       the WCS fit */

    cpl_msg_info(fctid,"Doing WCS on mosaic");
    (void)casu_getstds(casu_fits_get_ehu(ps.outmos),1,
                       ps.catpath,ps.catname,
                       vircam_sci_postproc_config.cdssearch,
                       vircam_sci_postproc_config.cacheloc,&stdscat,&status);
    (void)casu_matchstds(casu_tfits_get_table(ps.outmoscat),stdscat,300.0,
                         &tab,&status);
    (void)casu_platesol(casu_fits_get_ehu(ps.outmos),
                        casu_tfits_get_ehu(ps.outmoscat),tab,6,1,
                        &status);
    if (status != CASU_OK) {
        freetable(stdscat);
        freetable(tab);
        vircam_sci_postproc_tidy();
        return(-1);
    }
    freetable(tab);
    
    /* Grout the output catalogue (if requested) */

    if (vircam_sci_postproc_config.grout) {
        n = (int)cpl_frameset_get_size(ps.paws);
        if (n != 6) {
            cpl_msg_info(fctid,
                         "Only %" CPL_SIZE_FORMAT " paws. No grout possible",
                         (cpl_size)n);
        } else {
            cpl_msg_info(fctid,"Grouting catalogue");
            (void)vircam_grout(ps.outmoscat,ps.cats,ps.confs,&outmoscatg,
                               &status);
            freetfits(ps.outmoscat);
            ps.outmoscat = outmoscatg;
        }
    }

    /* Now do the photometry */

    cpl_msg_info(fctid,"Doing photometry on mosaic");
    (void)casu_matchstds(casu_tfits_get_table(ps.outmoscat),stdscat,300.0,
                         &tab,&status);
    freetable(stdscat);
    ps.tphottab = cpl_table_load(cpl_frame_get_filename(ps.phottab),1,0);
    vircam_pfits_get_filter(casu_fits_get_phu(ps.outmos),filt);
    matchstds = casu_tfits_wrap(tab,ps.outmoscat,NULL,NULL);
    casu_photcal_extinct(&(ps.outmos),&matchstds,&(ps.outmoscat),1,
                         filt,ps.tphottab,
                         vircam_sci_postproc_config.minphotom,
                         ps.schlf_n,ps.schlf_s,"EXPTIME","ESO TEL AIRM START",
                         vircam_sci_postproc_config.magerrcut,&status);
    cpl_table_save(tab,NULL,NULL,"mstds.fits",CPL_IO_DEFAULT);
    if (status != CASU_OK) {
        freetfits(matchstds);
        vircam_sci_postproc_tidy();
        return(-1);
    }
    freetfits(matchstds);

    /* Save the mosaic and its confidence map */

    cpl_msg_info(fctid,"Saving mosaic, confidence map and catalogue");
    (void)vircam_sci_postproc_save_image(ps.outmos,framelist,parlist,
                                         cpl_frameset_get_position(ps.paws,0),
                                         1); 
    (void)vircam_sci_postproc_save_image(ps.outmosconf,framelist,parlist,
                                         cpl_frameset_get_position(ps.paws,0),
                                         0); 
    (void)vircam_sci_postproc_save_cat(ps.outmoscat,framelist,parlist,
                                       cpl_frameset_get_position(ps.paws,0));

    /* Tidy up */

    vircam_sci_postproc_tidy();
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_image
    \par Purpose:
        Save the image products
    \par Description:
        An extension of an image product is saved here
    \par Language:
        C
    \param outim
        The casu_fits structure for the image to be saved
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \param template
        A template to be used in the cpl saving routine
    \param isim
        Set if this is the image. If not it's the confidence map 
    \retval CASU_OK if everything is ok
    \retval CASU_FATAL if something failed
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_postproc_save_image(casu_fits *outim, 
                                          cpl_frameset *framelist,
                                          cpl_parameterlist *parlist,
                                          cpl_frame *template, int isim) {
    char fname[VIRCAM_PATHSZ];
    cpl_frame *product_frame;
    cpl_propertylist *plist,*plist2;
    cpl_type type;
    int ptype;
    const char *fctid="vircam_sci_postproc_save_image";

    /* Get some information so we can set a file name */

    ptype = (isim ? 0 : 1);
    vircam_sci_product_name(cpl_frame_get_filename(template),ptype,
                            vircam_sci_postproc_config.prettynames,1,fname);

    /* Set up the file */

    if (access(fname,F_OK))
        remove(fname);
    product_frame = cpl_frame_new();
    cpl_frame_set_filename(product_frame,fname);
    if (isim) 
        cpl_frame_set_tag(product_frame,VIRCAM_PRO_MOSAIC);
    else 
        cpl_frame_set_tag(product_frame,VIRCAM_PRO_MOSAIC_CONF);
    cpl_frame_set_type(product_frame,CPL_FRAME_TYPE_IMAGE);
    cpl_frame_set_group(product_frame,CPL_FRAME_GROUP_PRODUCT);
    cpl_frame_set_level(product_frame,CPL_FRAME_LEVEL_FINAL);

    /* Set up the PHU header */

    plist = cpl_propertylist_duplicate(casu_fits_get_phu(outim));
    plist2 = cpl_propertylist_duplicate(casu_fits_get_ehu(outim));
    casu_merge_propertylists(plist,plist2);
    freepropertylist(plist2);
    vircam_dfs_set_product_primary_header(plist,product_frame,framelist,parlist,
                                          vircam_recipename,"PRO-1.15",
                                          template,1);

    /* Now save the data */

    if (isim) 
        type = CPL_TYPE_FLOAT;
    else
        type = CPL_TYPE_INT;
    if (cpl_image_save(casu_fits_get_image(outim),fname,type,
                       plist,CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product image extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }
    cpl_frameset_insert(framelist,product_frame);
    freepropertylist(plist);
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_cat
    \par Purpose:
        Save the catalogue products
    \par Description:
        An extension of a catalogue product is saved here
    \par Language:
        C
    \param outcat
        The casu_tfits structure for the catalogue to be saved
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \param template
        A template to be used in the cpl saving routine
    \retval CASU_OK if everything is ok
    \retval CASU_FATAL if something failed
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_postproc_save_cat(casu_tfits *outcat, 
                                        cpl_frameset *framelist,
                                        cpl_parameterlist *parlist,
                                        cpl_frame *template) {
    char fname[VIRCAM_PATHSZ];
    cpl_frame *product_frame;
    cpl_propertylist *plist;
    const char *fctid = "vircam_sci_postproc_save_cat";

    /* Get some information so we can set a file name */

    vircam_sci_product_name(cpl_frame_get_filename(template),2,
                            vircam_sci_postproc_config.prettynames,1,fname);

    /* Set up the file */

    if (access(fname,F_OK))
        remove(fname);
    product_frame = cpl_frame_new();
    cpl_frame_set_filename(product_frame,fname);
    cpl_frame_set_tag(product_frame,VIRCAM_PRO_OBJCAT_SCI);
    cpl_frame_set_type(product_frame,CPL_FRAME_TYPE_TABLE);
    cpl_frame_set_group(product_frame,CPL_FRAME_GROUP_PRODUCT);
    cpl_frame_set_level(product_frame,CPL_FRAME_LEVEL_FINAL);

    /* Set up the PHU header */

    plist = casu_tfits_get_phu(outcat);
    vircam_dfs_set_product_primary_header(plist,product_frame,framelist,parlist,
                                          vircam_recipename,"PRO-1.15",NULL,0);

    /* 'Save' the PHU image */                   

    if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,plist,
                       CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product PHU");
        cpl_frame_delete(product_frame);
        return(-1);
    }
    cpl_frameset_insert(framelist,product_frame);

    /* Now save the extension */

    plist = casu_tfits_get_ehu(outcat);
    vircam_dfs_set_product_exten_header(plist,product_frame,framelist,
                                        parlist,vircam_recipename,
                                        "PRO-1.15",template);
    if (cpl_table_save(casu_tfits_get_table(outcat),NULL,plist,fname,
                       CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product table extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_testfrm_1
    \par Purpose:
        Test a frame to make sure it is what you expect
    \par Description:
        The frame is checked to make sure it has the right number of 
        extensions and that each image/table is loadable.
    \par Language:
        C
    \param fr
        The input frame
    \param nextn_expected
        The number of extensions expected in each image/table file
    \param isimg
        TRUE if this is an image frame. Otherwise we check for tables
    \return
        The number of errors
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
    
static int vircam_sci_testfrm_1(cpl_frame *fr, int nextn_expected, int isimg) {
    int nextn,nerr,j;
    casu_fits *test;
    casu_tfits *testt;
    const char *fctid="vircam_sci_testfrm";

    /* Return immediately if given nonsense */

    if (fr == NULL)
        return(0);

    /* Test to see how many extensions there are and compare to see
       if it matches the number expected */

    nextn = cpl_frame_get_nextensions(fr);
    if (nextn != nextn_expected) {
        cpl_msg_error(fctid,"Frame %s has %" CPL_SIZE_FORMAT " extensions, expected %" CPL_SIZE_FORMAT "\n",
                      cpl_frame_get_filename(fr),(cpl_size)nextn,
                      (cpl_size)nextn_expected);
        return(1);
    }

    /* Test to see if you can load each of the extensions */

    nerr = 0;
    for (j = 1; j <= nextn; j++) {
        if (isimg) {
            test = casu_fits_load(fr,CPL_TYPE_FLOAT,j);
            if (test == NULL) {
                cpl_msg_error(fctid,
                              "Frame image %s[%" CPL_SIZE_FORMAT "] won't load\n",
                              cpl_frame_get_filename(fr),(cpl_size)j);
                nerr++;
                continue;
            }
            freefits(test);
        } else {
            testt = casu_tfits_load(fr,j);
            if (testt == NULL) {
                cpl_msg_error(fctid,
                              "Frame table %s[%" CPL_SIZE_FORMAT "] won't load\n",
                              cpl_frame_get_filename(fr),(cpl_size)j);
                nerr++;
                continue;
            }
            freetfits(testt);
        }
    }
    return(nerr);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_product_name
    \par Purpose:
        Set up an output product file name
    \par Description:
        An output file name is defined based on a template name,  a
        'predictable' name or a temporary file name.
    \par Language:
        C
    \param template
        If the pretty file name option is used or if this is a temporary
        file name, then the this will be used as a reference name.
    \param producttype
        The type of product you are writing out. These are enumerated above.
    \param nametype
        This is: 0 -> predictable names, 1 -> pretty names based on input
        file names or 2 -> temporary file names.
    \param fnumber
        If the predictable file names are going to be used, then this is 
        a number that is appended to a file name root.
    \param outfname
        The output file name
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_product_name(const char *template, int producttype, 
                                    int nametype, int fnumber, char *outfname) {
    const char *esonames[] = {"tile_","tile_conf_","tile_cat_"};
    const char *suffix[] = {"_tl","_tl_conf","_tl_cat"};
    char *fname,*bname,*dot;

    /* If the name type is the ESO predictable sort, then it's easy. Just
       use the esonames defined above and append the correct number */

    switch (nametype) {
    case 0:
        (void)sprintf(outfname,"%s%d.fits",esonames[producttype],fnumber);
        break;

    /* If this is a temporary file, then just append tmp_ to the template
       name and return that */

    case 2:
        fname = cpl_strdup(template);
        bname = basename(fname);
        (void)sprintf(outfname,"tmp_%s",bname);
        freespace(fname);
        break;

    /* Ok, we want a pretty name... */

    case 1:
        fname = cpl_strdup(template);
        bname = basename(fname);
        (void)sprintf(outfname,"%s",bname);
        dot = strrchr(outfname,'.');
        (void)sprintf(dot,"%s.fits",suffix[producttype]);
        freespace(fname);
        break;

    /* something else ?? */

    default:
        (void)strcpy(outfname,"");
        break;
    }
    return;
}  

/**@}*/

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_postproc_init
    \par Purpose:
        Initialise pointers in the global memory structure.
    \par Description:
        Initialise pointers in the global memory structure.
    \par Language:
        C
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_postproc_init(void) {
    ps.labels = NULL;
    ps.paws = NULL;
    ps.confs = NULL;
    ps.cats = NULL;
    ps.phottab = NULL;
    ps.tphottab = NULL;
    ps.catindex = NULL;
    ps.catpath = NULL;
    ps.catname = NULL;
    ps.schlf_n = NULL;
    ps.schlf_s = NULL;
    ps.outmos = NULL;
    ps.outmosconf = NULL;
    ps.outmoscat = NULL;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_postproc_tidy
    \par Purpose:
        Free allocated workspace in the global memory structure
    \par Description:
        Free allocated workspace in the global memory structure. 
    \par Language:
        C
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_postproc_tidy(void) {
    freespace(ps.labels);
    freeframeset(ps.paws);
    freeframeset(ps.confs);
    freeframeset(ps.cats);
    freeframe(ps.phottab);
    freetable(ps.tphottab);
    freeframe(ps.catindex);
    freespace(ps.catpath);
    freespace(ps.catname);
    freeframe(ps.schlf_n);
    freeframe(ps.schlf_s);
    freefits(ps.outmos);
    freefits(ps.outmosconf);
    freetfits(ps.outmoscat);
}

/*

$Log: not supported by cvs2svn $
Revision 1.1  2013/10/15 17:01:44  jim
new entry


*/
