/* $Id: vircam_dfs.c,v 1.43 2013-10-15 16:32:50 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:32:50 $
 * $Revision: 1.43 $
 * $Name: not supported by cvs2svn $
 */


/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <cpl.h>

#include "vircam_dfs.h"
#include "vircam_utils.h"

#define PACKSZ 1024
#define DELEXTN "^(ESO.INS |ESO.OBS |ESO.TEL |INSTRUME|PIPEFILE|ORIGIN" \
                "|TELESCOP|OBJECT|RA|DEC|EPOCH|EQUINOX|DATAMD5|RADECSYS" \
                "|RADESYS|HDRVER|DATE-OBS|MJD-OBS|UTC|LST|PI-COI|OBSERVER).*$"

#define REGFIX "^(ORIGIN|TELESCOP|INSTRUME|OBJECT|RA|DEC|EPOCH|EQUINOX" \
    "|RADECSYS|RADESYS|DATE-OBS|MJD-OBS|UTC|LST|PI-COI|OBSERVER)$"

/**
   \defgroup vircam_dfs vircam_dfs
   \ingroup supportroutines

   \brief
   These are methods for manipulating keywords in data product headers

   \author
   Jim Lewis, CASU

*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Rename the reference frame keyword (RADECSYS -> RADESYS)
  @param    list    The propertylist to operate on
  @return   The resulting error code
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code vircam_dfs_rename_ref_frame(cpl_propertylist * list) {
    if (!list) return CPL_ERROR_NULL_INPUT;

    if (cpl_propertylist_has(list, "RADECSYS")) {
        if (cpl_propertylist_has(list, "RADESYS"))
            cpl_propertylist_erase(list, "RADECSYS");
        else {
            char * radec_val = cpl_strdup(cpl_propertylist_get_string(
                    list, "RADECSYS"));
            char * radec_com = cpl_strdup(cpl_propertylist_get_comment(
                    list, "RADECSYS"));
            cpl_propertylist_erase(list, "RADECSYS");
            cpl_propertylist_update_string(list, "RADESYS", radec_val);
            cpl_propertylist_set_comment(list, "RADESYS", radec_com);
            cpl_free(radec_val);
            cpl_free(radec_com);
        }
    }
    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_dfs_set_groups
    \par Purpose:
        Set the group of each frame in a frameset as RAW or CALIB according
        to DO.CATG tag given to it.
    \par Description:
        A frameset is given. Each frame in the frameset is classified as
        RAW or CALIB depending upon the dfs tag it's been given.
    \par Language:
        C
    \param set     
        The input frameset
    \retval CASU_OK 
        If everything is ok
    \retval CASU_FATAL 
        If NULL frameset
    \author
        Jim Lewis, CASU

 */
/*---------------------------------------------------------------------------*/

extern int vircam_dfs_set_groups(cpl_frameset *set) {
    cpl_frame *cur_frame;
    const char *tag;
    const char *fctid = "vircam_dfs_set_groups";
    int nframes,i;

    /* Check entries */

    if (set == NULL) 
        return(CASU_FATAL); 
    
    /* Initialize */

    nframes = cpl_frameset_get_size(set);

    /* Loop on frames */

    for (i = 0; i < nframes; i++) {
        cur_frame = cpl_frameset_get_position(set,i);
        tag = cpl_frame_get_tag(cur_frame);
       
        /* RAW frames */

        if (!strcmp(tag,VIRCAM_DARKCUR_RAW)  ||
            !strcmp(tag,VIRCAM_DARK_RAW)  ||
            !strcmp(tag,VIRCAM_NOISE_FLAT_RAW) ||
            !strcmp(tag,VIRCAM_NOISE_DARK_RAW) ||
            !strcmp(tag,VIRCAM_DOME_RAW) ||
            !strcmp(tag,VIRCAM_SCI_OBJECT_RAW) || 
            !strcmp(tag,VIRCAM_SCI_OBJECT_EXT_RAW) || 
            !strcmp(tag,VIRCAM_OFFSET_SKY_RAW) ||
            !strcmp(tag,VIRCAM_TEST_SCIENCE_RAW) || 
            !strcmp(tag,VIRCAM_LIN_DOME_RAW) ||
            !strcmp(tag,VIRCAM_LIN_DOME_CHECK) ||
            !strcmp(tag,VIRCAM_LIN_DARK_RAW) ||
            !strcmp(tag,VIRCAM_LIN_DARK_CHECK) ||
            !strcmp(tag,VIRCAM_ILLUM_RAW) ||
            !strcmp(tag,VIRCAM_RESET_RAW) ||
            !strcmp(tag,VIRCAM_STD_OBJECT_RAW) ||
            !strcmp(tag,VIRCAM_TWI_RAW) ||
            !strcmp(tag,VIRCAM_PRO_OBJCAT_SCI) ||
            !strcmp(tag,VIRCAM_PRO_CONF_SCI) || 
            !strcmp(tag,VIRCAM_PRO_JITTERED_SCI))
            cpl_frame_set_group(cur_frame,CPL_FRAME_GROUP_RAW);

        /* CALIB frames */

        else if (!strcmp(tag,VIRCAM_CAL_RESET) ||
                 !strcmp(tag,VIRCAM_CAL_DARK) ||
                 !strcmp(tag,VIRCAM_CAL_DOME_FLAT) ||
                 !strcmp(tag,VIRCAM_CAL_TWILIGHT_FLAT) ||
                 !strcmp(tag,VIRCAM_CAL_CONF) ||
                 !strcmp(tag,VIRCAM_CAL_BPM) ||
                 !strcmp(tag,VIRCAM_CAL_FRINGE) ||
                 !strcmp(tag,VIRCAM_CAL_CHANTAB) ||
                 !strcmp(tag,VIRCAM_CAL_READGAINFILE) ||
                 !strcmp(tag,VIRCAM_CAL_CHANTAB_INIT) ||
                 !strcmp(tag,VIRCAM_CAL_PHOTTAB) ||
                 !strcmp(tag,VIRCAM_CAL_STDTAB) ||
                 !strcmp(tag,VIRCAM_CAL_OBJCAT) ||
                 !strcmp(tag,VIRCAM_CAL_MSTDTAB) ||
                 !strcmp(tag,VIRCAM_CAL_2MASS) ||
                 !strcmp(tag,VIRCAM_CAL_SCHL_N) ||
                 !strcmp(tag,VIRCAM_CAL_SCHL_S) ||
                 !strcmp(tag,VIRCAM_CAL_REFCAT) ||
                 !strcmp(tag,VIRCAM_CAL_SKY) ||
                 !strcmp(tag,VIRCAM_CAL_OBJMASK) ||
                 !strcmp(tag,VIRCAM_REF_RESET) ||
                 !strcmp(tag,VIRCAM_REF_DARK) ||
                 !strcmp(tag,VIRCAM_REF_DOME_FLAT) ||
                 !strcmp(tag,VIRCAM_REF_TWILIGHT_FLAT))
            cpl_frame_set_group(cur_frame,CPL_FRAME_GROUP_CALIB);
        else 
            cpl_msg_info(fctid,"No such tag as %s in frame %s",tag,
                         cpl_frame_get_filename(cur_frame));
    }
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_dfs_set_product_primary_header
    \par Purpose:
        Set the dfs product header keywords in the primary header
    \par Description:
        Set the dfs product header keywords in the primary header
    \par Language:
        C
    \param plist     
        The propertylist to which the keywords should be appended
    \param frame
        The output product frame
    \param frameset
        The input recipe frameset
    \param parlist
        The parameter list of the input recipe
    \param recipeid
        The name of the input recipe
    \param dict
        The dictionary ID
    \param inherit
        The frame from which you want to inherit useful header information
    \param synch
        If set, then the timing information that has been added by the
        relevant cpl_ routine can be overwritten by the timing information
        from the original propertylist
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_dfs_set_product_primary_header(cpl_propertylist *plist,
                                                  cpl_frame *frame,
                                                  cpl_frameset *frameset,
                                                  cpl_parameterlist *parlist,
                                                  char *recipeid, 
                                                  const char *dict,
                                                  cpl_frame *inherit,
                                                  int synch) {
    const char *fctid="vircam_dfs_set_product_primary_header";
    char package[PACKSZ];
    cpl_propertylist *pcopy;

    /* Get the package name and version */

    (void)snprintf(package,PACKSZ,"%s/%s",PACKAGE,PACKAGE_VERSION);

    /* Make a copy of the input propertylist so that we can restore the 
       time and position dependent headers back to the way they were before
       cpl_dfs_setup_product_header worked on it. */

    pcopy = cpl_propertylist_duplicate(plist);

    /* Add the data-flow keywords */

    if (cpl_dfs_setup_product_header(plist,frame,frameset,parlist,recipeid,
                                     package,dict,inherit) != CPL_ERROR_NONE) {
        cpl_msg_warning(fctid,
                        "Problem with the product primary DFS compliance -- %s",
                        cpl_error_get_message());
        cpl_error_reset();
    }

    /* Now fix the time/position information back to the way it was */

    if (synch) 
        cpl_propertylist_copy_property_regexp(plist,
                                              (const cpl_propertylist *)pcopy,
                                              REGFIX,0);
    cpl_propertylist_erase_regexp(plist, "HDRVER", 0);
    vircam_dfs_rename_ref_frame(plist);
    cpl_propertylist_delete(pcopy);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_dfs_set_product_exten_header
    \par Purpose:
        Set the dfs product header keywords in the primary header
    \par Description:
        Set the dfs product header keywords in the primary header
    \par Language:
        C
    \param plist     
        The propertylist to which the keywords should be appended
    \param frame
        The output product frame
    \param frameset
        The input recipe frameset
    \param parlist
        The parameter list of the input recipe
    \param recipeid
        The name of the input recipe
    \param dict
        The dictionary ID
    \param inherit
        The frame from which you want to inherit all the useful header info
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_dfs_set_product_exten_header(cpl_propertylist *plist,
                                                cpl_frame *frame,
                                                cpl_frameset *frameset,
                                                cpl_parameterlist *parlist,
                                                char *recipeid, 
                                                const char *dict,
                                                cpl_frame *inherit) {
    const char *fctid="vircam_dfs_set_product_exten_header";
    char package[PACKSZ];

    /* Get the package name and version */

    (void)snprintf(package,PACKSZ,"%s/%s",PACKAGE,PACKAGE_VERSION);

    /* Add the data-flow keywords */

    if (cpl_dfs_setup_product_header(plist,frame,frameset,parlist,recipeid,
                                     package,dict,inherit) != CPL_ERROR_NONE) {
        cpl_msg_warning(fctid,
                        "Problem with the product extn DFS compliance -- %s",
                        cpl_error_get_message());
        cpl_error_reset();
    }

    /* Remove the stuff that isn't supposed to be here */

    cpl_propertylist_erase_regexp(plist,DELEXTN,0);

    /* ensure RADESYS & EQUINOX are inherited from PHU, if necessary */
    cpl_errorstate state = cpl_errorstate_get();
    cpl_wcs * wcs = cpl_wcs_new_from_propertylist(plist);
    if (wcs) {
        cpl_propertylist_update_bool(plist, "INHERIT", 1);
        cpl_propertylist_set_comment(plist, "INHERIT",
                                     "denotes the INHERIT keyword convention");
        cpl_wcs_delete(wcs);
    }
    cpl_errorstate_set(state);

}

/**@}*/

/*

$Log: not supported by cvs2svn $
Revision 1.42  2010/09/10 11:22:44  jim
Added VIRCAM_CAL_SKY entry

Revision 1.41  2010/01/31 18:59:54  jim
Removed references to master sky

Revision 1.40  2009/09/22 12:27:01  jim
Added entry for raw offset sky exposures

Revision 1.39  2009/09/09 09:44:54  jim
Modified in final ditch effort to get headers right...

Revision 1.38  2009/03/02 20:00:10  jim
Uses new version of cpl_dfs_set_product_header

Revision 1.37  2008/11/02 14:33:11  jim
Modified vircam_dfs_set_product_primary_header so that critical positional
and timing data are not overwritten

Revision 1.36  2008/10/21 08:37:15  jim
Added VIRCAM_PRO_SKYMAP_SCI, VIRCAM_CAL_SKY and VIRCAM_CAL_SKY_MASK

Revision 1.35  2008/10/07 11:21:56  jim
Fixed regexp again...

Revision 1.34  2008/09/30 11:34:14  jim
modified definition of DELEXTN

Revision 1.33  2008/09/29 11:20:17  jim
Modified to keep PRO keywords in the primary header. Extended the list of
items to be removed from the extension header.

Revision 1.32  2007/11/20 09:37:56  jim
Added VIRCAM_LIN_DOME_CHECK and VIRCAM_LIN_DARK_CHECK

Revision 1.31  2007/10/19 06:55:06  jim
Modifications made to use new method for directing the recipes to the
standard catalogues using the sof

Revision 1.30  2007/08/23 09:00:59  jim
Removed annoying error messages in the product header routines which used
to tell you if it couldn't find stuff to erase.

Revision 1.29  2007/04/23 12:50:22  jim
Added VIRCAM_TEST_SCIENCE_RAW

Revision 1.28  2007/04/04 10:34:42  jim
Complete rewrite of the dfs tags

Revision 1.27  2007/03/23 10:53:22  jim
Fixed little documentation errors

Revision 1.26  2007/03/01 12:42:41  jim
Modified slightly after code checking

Revision 1.25  2007/02/15 11:53:45  jim
Added type CHANNEL_TABLE_INIT

Revision 1.24  2007/02/05 13:54:00  jim
Added REFERENCE tags

Revision 1.23  2007/01/17 23:53:43  jim
Modified to remove INS, OBS and TEL structures from extension headers

Revision 1.22  2006/12/06 12:58:57  jim
Added MASTER_FRINGE

Revision 1.21  2006/11/28 20:57:05  jim
Added ILLUM types

Revision 1.20  2006/10/05 09:22:59  jim
Small modifications to a couple of cpl calls to bring them into line with
cpl v3.0

Revision 1.19  2006/07/04 09:19:04  jim
replaced all sprintf statements with snprintf

Revision 1.18  2006/06/14 14:13:58  jim
fixed minor doc problem

Revision 1.17  2006/06/13 21:26:02  jim
Added VIRCAM_CALIB_READNOISE_FILE

Revision 1.16  2006/05/24 13:33:56  jim
Added VIRCAM_PHOTTAB

Revision 1.15  2006/05/18 12:32:23  jim
Added channel table and object catalogue to calib list

Revision 1.14  2006/04/27 09:46:27  jim
Modified to conform to new dictionary

Revision 1.13  2006/04/26 09:39:35  jim
Fixes the pipeline id to the header

Revision 1.12  2006/04/25 13:47:59  jim
Modified calls to vircam_dfs_set_product_*_header routines

Revision 1.11  2006/04/20 11:19:22  jim
*** empty log message ***

Revision 1.10  2006/03/22 13:14:24  jim
Sorted data categories

Revision 1.9  2006/02/27 13:53:32  jim
Added new categories to vircam_dfs_set_groups

Revision 1.8  2006/02/18 11:45:01  jim
Added vircam_dfs_set_product_primary_header and
vircam_dfs_set_product_exten_header

Revision 1.7  2006/01/23 10:30:49  jim
Mainly documentation mods

Revision 1.6  2005/12/14 22:17:33  jim
Updated docs

Revision 1.5  2005/11/29 14:56:36  jim
Added a few new categories

Revision 1.4  2005/11/25 09:56:14  jim
Tidied up some more documentation

Revision 1.3  2005/11/03 13:28:48  jim
All sorts of changes to tighten up error handling

Revision 1.2  2005/09/20 15:07:46  jim
Fixed a few bugs and added a few things

Revision 1.1.1.1  2005/08/05 08:29:09  jim
Initial import


*/
