/* $Id: vircam_tilesky_minus.c,v 1.1 2013-10-15 16:30:07 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:30:07 $
 * $Revision: 1.1 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cpl.h>
#include <math.h>
#include <string.h>

#include <casu_utils.h>
#include <casu_mods.h>
#include <casu_mask.h>
#include <casu_stats.h>
#include <casu_wcsutils.h>

#include "vircam_sky.h"
#include "vircam_mods.h"
#include "vircam_dfs.h"
#include "vircam_pfits.h"

typedef struct {
    int pawposition;
    int nimages;
    casu_fits **infiles;
    float **datas;
    unsigned char **masks;
    float *master_xsky;
    unsigned char *master_xsky_bpm;
    float **xsky_variant;
    unsigned char **xsky_variant_bpm;
    float *zeros;
    float *dx;
    float *dy;
} jpos;
static void combine_master(int nfiles, int stat, int npts, float **datas, 
                           unsigned char **masks, float **skyout, 
                           unsigned char **skyout_bpm);
static void combine_variant(int stat, int npts, jpos *jp, int skipme);
static void combine_final(int nxskys, int stat, int npts, float **xskys,
                          unsigned char **xsky_masks, float *data, 
                          float *master_sky, unsigned char *master_bpm);
static void masksky_zeros(jpos *jp, int njit, int npts);
static void domed(float *buf, int n, float *val);
static jpos *jpos_init(int njit);
static void jpos_alloc(casu_fits **infiles, int nfiles, int pawpos, jpos *jp);
static void jpos_free(jpos *jp, int njit);

extern int vircam_tilesky_minus(casu_fits **infiles, casu_fits *conf, 
                                casu_fits *objmaskfits, int nfiles, 
                                casu_fits **skyout, int *status) {
    int i, nx_mask = 0,ind,j,nx,ny,kind,njit,*confdata = NULL,jx,jy,ix,iy,k;
    int npts,*opm = NULL;
    jpos *jp=NULL;
    cpl_wcs *wcsmask,*wcsimg;
    cpl_image *im,*skyim;
    cpl_propertylist *plist;
    unsigned char **xsky_masks,*master_sky_lev2_bpm;
    double *xin = NULL,*yin = NULL,*xout = NULL,*yout = NULL,*dx = NULL,*dy = NULL,ddx,ddy;
    float **xskys,*master_sky_lev2,val,sig;
    const char *fctid = "vircam_tilesky_minus";

    /* Inherited status */

    *skyout = NULL;
    if (*status != CASU_OK)
        return(*status);

    /* If there aren't any images, then get out of here */

    if (nfiles == 0) {
        cpl_msg_error(fctid,"Sky correction impossible. No science frames");
        return(CASU_FATAL);
    }

    /* Initialise the jpos structure. There is one for each jitter position.
       Each jitter position has a maximum number of observations which is
       the number of jitter sequences. This is nominally 6, but will 
       sometimes be 3 if we are observing in stripes or something else
       if the tile is incomplete. */

    (void)vircam_pfits_get_njsteps(casu_fits_get_phu(infiles[0]),&njit);
    jp = jpos_init(njit);
    for (i = 1; i <= njit; i++) 
        jpos_alloc(infiles,nfiles,i,jp+i-1);
    nx = cpl_image_get_size_x(casu_fits_get_image(infiles[0]));
    ny = cpl_image_get_size_y(casu_fits_get_image(infiles[0]));
    npts = nx*ny;

    /* Get the wcs descriptor for the mask (if it exists) */

    if (objmaskfits != NULL) {
        nx_mask = cpl_image_get_size_x(casu_fits_get_image(objmaskfits));
        wcsmask = cpl_wcs_new_from_propertylist(casu_fits_get_ehu(objmaskfits));
        opm = cpl_image_get_data_int(casu_fits_get_image(objmaskfits));

        /* We need this space so that we can work out which part of the 
           input mask is relevant for each input image */
    
        xin = cpl_malloc(npts*sizeof(double));
        yin = cpl_malloc(npts*sizeof(double));
        xout = cpl_malloc(npts*sizeof(double));
        yout = cpl_malloc(npts*sizeof(double));
        dx = cpl_malloc(nfiles*sizeof(double));
        dy = cpl_malloc(nfiles*sizeof(double));

        /* Initialise the input xy arrays */

        ind = 0;
        for (j = 0; j < ny; j++) {
            for (i = 0; i < nx; i++) {
                xin[ind] = (double)(i+1);
                yin[ind++] = (double)(j+1);
            }
        }

        /* Define an output grid for the first image and then offsets for
           the subsequent images */

        ddx = 1.0;
        ddy = 1.0;
        for (i = 0; i < nfiles; i++) {
            wcsimg = cpl_wcs_new_from_propertylist(casu_fits_get_ehu(infiles[i]));
            if (i == 0) 
                casu_xytoxy_list(wcsimg,wcsmask,npts,xin,yin,xout,yout);
            casu_xytoxy_list(wcsimg,wcsmask,1,&ddx,&ddy,dx+i,dy+i);
            dx[i] -= xout[0];
            dy[i] -= yout[0];
            cpl_wcs_delete(wcsimg);
        }
        cpl_wcs_delete(wcsmask);
    }

    /* Loop for each file and get a reference for the data array. Get
       some space for a mask that will cover both the zeroed pixels in the
       confidence maps and any object mask that might be used */

    for (j = 0; j < njit; j++) {
        for (i = 0; i < jp[j].nimages; i++) {
            im = casu_fits_get_image(jp[j].infiles[i]);
            jp[j].datas[i] = cpl_image_get_data_float(im);
            jp[j].masks[i] = cpl_calloc(npts,sizeof(unsigned char));
            if (i == 0 && j == 0) 
                confdata = cpl_image_get_data_int(casu_fits_get_image(conf));

            /* Now loop for each pixel. If the confidence map is zero here then
               mark this as a bad pixel. If not, then look at the master mask
               and see which is its closest pixel to the current one. NB: the 
               -0.5 in the index is a combination of +0.5 to do a nint and a -1
               to take account of the fact that arrays start from index zero. */

            for (jy = 0; jy < ny; jy++) {
                for (jx = 0; jx < nx; jx++) {
                    ind = jy*nx + jx;
                    if (confdata[ind] == 0) {
                        jp[j].masks[i][ind] = 1;
                    } else if (objmaskfits != NULL) {
                        kind = (int)(yout[ind] + dy[i] - 0.5)*nx_mask + 
                            (int)(xout[ind] + dx[i] - 0.5);
                        jp[j].masks[i][ind] = opm[kind];
                    }
                }
            }
        }
    }

    /* Do an intermediate tidy */

    if (objmaskfits != NULL) {
        freespace(xin);
        freespace(yin);
        freespace(xout);
        freespace(yout);
        freespace(dx);
        freespace(dy);
    }

    /* Offset the DC level for all images together */

    masksky_zeros(jp,njit,npts);

    /* Now loop through all the positions again and form the xskys and
       all their variants. The 'master' is an xsky that is formed from
       all the images. The variant xskys are formed from all but one
       of the input images (removing one image from the list each time). */

    for (i = 0; i < njit; i++) {

        /* Form the master xsky for this position */

        combine_master(jp[i].nimages,0,npts,jp[i].datas,jp[i].masks,
                       &(jp[i].master_xsky),&(jp[i].master_xsky_bpm));

        /* Form the variant xskys for this position */

        for (j = 0; j < jp[i].nimages; j++) 
            combine_variant(0,npts,jp+i,j);
    }

    /* Form master level 2 sky */

    xskys = cpl_malloc(njit*sizeof(float *));
    xsky_masks = cpl_malloc(njit*sizeof(unsigned char *));
    for (i = 0; i < njit; i++) {
        xskys[i] = jp[i].master_xsky;
        xsky_masks[i] = jp[i].master_xsky_bpm;
    }
    combine_master(njit,0,npts,xskys,xsky_masks,&master_sky_lev2,
                   &master_sky_lev2_bpm);

    /* Wrap the sky and reject the pixels that had no sky contribution */

    skyim = cpl_image_wrap_float(nx,ny,master_sky_lev2);
    for (i = 0; i < npts; i++) {
        iy = i/nx + 1;
        ix = i - (iy-1)*nx + 1;
        if (master_sky_lev2_bpm[i])
            cpl_image_reject(skyim,(cpl_size)ix,(cpl_size)iy);
    }
    *skyout = casu_fits_wrap(skyim,infiles[0],NULL,NULL);
    plist = casu_fits_get_ehu(*skyout);
    if (objmaskfits != NULL) {
        cpl_propertylist_update_string(plist,"ESO DRS MASKUSED",
                                       casu_fits_get_filename(objmaskfits));
        cpl_propertylist_set_comment(plist,"ESO DRS MASKUSED",
                                     "Object masked used to make sky");
    }
    cpl_propertylist_update_string(plist,"ESO DRS SKYALGO","tilesky_minus");
    cpl_propertylist_set_comment(plist,"ESO DRS SKYALGO",
                                 "Sky estimation algorithm");
    casu_prov(plist,infiles,nfiles,1);

    /* Fill in the bits with no sky contribution */

    casu_inpaint(*skyout,64,status);

    /* OK, now loop through each of the jitter positions again and 
       form the variant level 2 sky and subtract it off the data frames */

    for (j = 0; j < njit; j++) {
        for (i = 0; i < njit; i++) {
            xskys[i] = jp[i].master_xsky;
            xsky_masks[i] = jp[i].master_xsky_bpm;
        }
        for (i = 0; i < jp[j].nimages; i++) {
            xskys[j] = jp[j].xsky_variant[i];
            xsky_masks[j] = jp[j].xsky_variant_bpm[i];
            combine_final(njit,0,npts,xskys,xsky_masks,jp[j].datas[i],
                          master_sky_lev2,master_sky_lev2_bpm);

            /* Readjust the DC to be what it was before sky subtraction */

            casu_qmedsig(jp[j].datas[i],jp[j].masks[i],(long)npts,3.0,1,
                         -1000.0,65535.0,&val,&sig);    
            val = jp[j].zeros[i] - val;
            for (k = 0; k < npts; k++)
                jp[j].datas[i][k] += val;
        }
    }

    /* Tidy up */

    freespace(xskys);
    freespace(xsky_masks);
    freespace(master_sky_lev2_bpm);
    jpos_free(jp,njit);
    return(CASU_OK);
}

static void combine_master(int nfiles, int stat, int npts, float **datas, 
                           unsigned char **masks, float **skyout, 
                           unsigned char **skyout_bpm) {
    int i,j,n;
    float *buf;

    /* Get workspace for output arrays */

    *skyout = cpl_malloc(npts*sizeof(float));
    *skyout_bpm = cpl_malloc(npts*sizeof(unsigned char));
    buf = cpl_malloc(nfiles*sizeof(float));

    /* Loop now for each input pixel and form the median */

    for (j = 0; j < npts; j++) {
        n = 0;
        for (i = 0; i < nfiles; i++) {
            if (masks[i][j] == 0) 
                buf[n++] = datas[i][j];
        }
        if (n == 0) {
            (*skyout)[j] = 0.0;
            (*skyout_bpm)[j] = 1;
        } else {
            if (stat == 0) 
                domed(buf,n,(*skyout)+j);
            else 
                (*skyout)[j] = casu_mean(buf,NULL,n);
            (*skyout_bpm)[j] = 0;
        }
    }

    /* Free some workspace and get out of here */

    freespace(buf);
    return;
}

static void combine_variant(int stat, int npts, jpos *jp, int skipme) {
    int i,j,n,nfiles;
    float *buf,**datas,*skyout,*master_sky;
    unsigned char **masks,*skyout_bpm,*master_sky_bpm;

    /* Create some shortcuts */

    datas = jp->datas;
    masks = jp->masks;
    master_sky = jp->master_xsky;
    master_sky_bpm = jp->master_xsky_bpm;
    nfiles = jp->nimages;

    /* Get workspace for output arrays */

    skyout = cpl_malloc(npts*sizeof(float));
    jp->xsky_variant[skipme] = skyout;
    skyout_bpm = cpl_malloc(npts*sizeof(unsigned char));
    jp->xsky_variant_bpm[skipme] = skyout_bpm;
    buf = cpl_malloc(nfiles*sizeof(float));

    /* If there is only one file in this group then just replace the 
       variant with the master and get out of here */

    if (nfiles == 1) {
        memmove(skyout,jp->master_xsky,npts*sizeof(float));
        memmove(skyout_bpm,jp->master_xsky_bpm,npts*sizeof(unsigned char));
        freespace(buf);
        return;
    }

    /* Loop now for each input pixel and form the median */

    for (j = 0; j < npts; j++) {
        n = 0;

        /* If the current pixel was masked out in the skipped image, then
           there isn't any point in recalculating the mean value again */

        if (masks[skipme][j]) {
            skyout[j] = master_sky[j];
            skyout_bpm[j] = master_sky_bpm[j];
            continue;
        }

        /* Recalculate */

        for (i = 0; i < nfiles; i++) {
            if (masks[i][j] == 0 && i != skipme) 
                buf[n++] = datas[i][j];
        }
        if (n == 0) {
            skyout[j] = 0.0;
            skyout_bpm[j] = 1;
        } else {
            if (stat == 0) 
                domed(buf,n,skyout+j);
            else 
                skyout[j] = casu_mean(buf,NULL,n);
            skyout_bpm[j] = 0;
        }
    }

    /* Free some workspace and get out of here */

    freespace(buf);
    return;
}

static void combine_final(int nxskys, int stat, int npts, float **xskys,
                          unsigned char **xsky_masks, float *data, 
                          float *master_sky, unsigned char *master_bpm) {
    int i,j,n;
    float *buf,val;

    /* Get some memory */

    buf = cpl_malloc(nxskys*sizeof(float));

    /* Loop now for each input pixel */

    for (j = 0; j < npts; j++) {

        /* If the master bpm is flagged, then this is an interpolated point
           and there's no reason to recalculate */

        if (master_bpm[j]) {
            val = master_sky[j];
        } else {

            /* Loop for each file. If the current pixel in the current file
               was flagged, then the sky for that pixel will be the same as
               the general sky we did before. If not, then remove that file's
               pixel and recalculate the sky */

            n = 0;
            for (i = 0; i < nxskys; i++) {
                if (xsky_masks[i][j] != 0)
                    continue;
                buf[n++] = xskys[i][j];
            }
            if (n == 0) {
                val = master_sky[j];
            } else {
                if (stat == 0) 
                    domed(buf,n,&val);
                else 
                    val = casu_mean(buf,NULL,n);
            }
        }

        /* Now correct this pixel in the input image */

        data[j] -= val;
    }

    /* Free some workspace and get out of here */

    freespace(buf);
    return;
}


static jpos *jpos_init(int njit) {
    jpos *jp;
    int i;

    jp = cpl_malloc(njit*sizeof(jpos));
    for (i = 0; i < njit; i++) {
        jp[i].pawposition = i + 1;
        jp[i].nimages = 0;
        jp[i].infiles = NULL;
        jp[i].datas = NULL;
        jp[i].masks = NULL;
        jp[i].master_xsky = NULL;
        jp[i].master_xsky_bpm = NULL;
        jp[i].xsky_variant = NULL;
        jp[i].xsky_variant_bpm = NULL;
        jp[i].zeros = NULL;
        jp[i].dx = NULL;
        jp[i].dy = NULL;
    }
    return(jp);
}

static void jpos_alloc(casu_fits **infiles, int nfiles, int pawpos, jpos *jp) {
    int i,ji;

    /* First the paw position */

    jp->pawposition = pawpos;
    
    /* Now loop through the files and find which ones are at this position */

    jp->nimages = 0;
    jp->infiles = cpl_malloc(6*sizeof(casu_fits *));
    for (i = 0; i < nfiles; i++) {
        (void)vircam_pfits_get_jitteri(casu_fits_get_phu(infiles[i]),&ji);
        if (ji != pawpos) 
            continue;
        jp->infiles[jp->nimages] = infiles[i];
        jp->nimages += 1;
    }
    jp->infiles = cpl_realloc(jp->infiles,jp->nimages*sizeof(casu_fits *));

    /* Right, safely through all the hard bits. Finish allocating memory */

    jp->datas = cpl_calloc(jp->nimages,sizeof(float *));
    jp->masks = cpl_calloc(jp->nimages,sizeof(unsigned char *));
    jp->master_xsky = NULL;
    jp->master_xsky_bpm = NULL;
    jp->xsky_variant = cpl_calloc(jp->nimages,sizeof(float *));
    jp->xsky_variant_bpm = cpl_calloc(jp->nimages,sizeof(unsigned char *));
    jp->zeros = cpl_calloc(jp->nimages,sizeof(float));
    jp->dx = cpl_calloc(jp->nimages,sizeof(float));
    jp->dy = cpl_calloc(jp->nimages,sizeof(float));
}
    
static void jpos_free(jpos *jp, int njit) {
    int i,j;

    for (i = 0; i < njit; i++) {
        freespace(jp[i].infiles);
        freespace(jp[i].master_xsky);
        freespace(jp[i].master_xsky_bpm);
        freespace(jp[i].zeros);
        for (j = 0; j < jp[i].nimages; j++) {
            freespace(jp[i].masks[j]);
            freespace(jp[i].xsky_variant[j]);
            freespace(jp[i].xsky_variant_bpm[j]);
        }
        freespace(jp[i].datas);
        freespace(jp[i].masks);
        freespace(jp[i].xsky_variant);
        freespace(jp[i].xsky_variant_bpm);
        freespace(jp[i].dx);
        freespace(jp[i].dy);
    }
    freespace(jp);
}   

static void masksky_zeros(jpos *jp, int njit, int npts) {
    int i,j,k,nf;
    float sig,medval,off,*zeros;
    
    /* How many files total are there? */
    
    nf = 0;
    for (j = 0; j < njit; j++)
        nf += jp[j].nimages;

    /* Get some workspace */

    zeros = cpl_malloc(nf*sizeof(float));

    /* Loop for each input image and get the background median */

    nf = 0;
    for (j = 0; j < njit; j++) {
        for (i = 0; i < jp[j].nimages; i++) {
            casu_qmedsig(jp[j].datas[i],jp[j].masks[i],(long)npts,3.0,1,
                         -1000.0,65535.0,zeros+nf,&sig);
            jp[j].zeros[i] = zeros[nf];
            nf++;
        }
    }       

    /* Get the median value of the array */

    medval = casu_med(zeros,NULL,nf);

    /* Now work out the offset for each image */

    for (j = 0; j < njit; j++) {
        for (i = 0; i < jp[j].nimages; i++) {
            off = medval - jp[j].zeros[i];
            for (k = 0; k < npts; k++) 
                jp[j].datas[i][k] += off;
        }
    }
    freespace(zeros);
}

static void domed(float *buf, int n, float *val) {
    int is_even,n2;

    is_even = ! (n & 1);
    casu_sort(&buf,n,1);
    n2 = n/2;
    if (is_even) {
        *val = 0.5*(buf[n2-1] + buf[n2]);
    } else {
        if (n <= 5) 
            *val = buf[n2];
        else
            *val = 0.5*buf[n2] + 0.25*(buf[n2-1] + buf[n2+1]);
    }
}

/*

$Log: not supported by cvs2svn $

*/
